from datetime import date
from enum import Enum

import requests
from pydantic import BaseModel

from stackit_cost_monitoring.auth import Auth


STACKIT_API_BASE_URL = 'https://cost.api.stackit.cloud/v3'


class CostApiException(Exception):
    pass


class CostApiItem(BaseModel):
    customerAccountId: str
    projectId: str
    projectName: str
    totalCharge: float
    totalDiscount: float


class CostApiTimePeriod(BaseModel):
    end: date
    start: date


class CostApiDetails(BaseModel):
    charge: float
    discount: float
    quantity: int
    timePeriod: CostApiTimePeriod


class CostApiItemWithDetails(CostApiItem):
    reportData: list[CostApiDetails]


class CostApiDepth(Enum):
    PROJECT = 'project'
    SERVICE = 'service'
    AUTO = 'auto'


class CostApiGranularity(Enum):
    NONE = 'none'
    DAILY = 'daily'
    WEEKLY = 'weekly'
    MONTHLY = 'monthly'
    YEARLY = 'yearly'


class CostApi:
    def __init__(self, auth: Auth):
        self._auth = auth

    def get_project_costs(
            self,
            customer_account_id: str,
            project_id: str,
            from_date: date,
            to_date: date,
            depth: CostApiDepth = CostApiDepth.AUTO,
            granularity: CostApiGranularity = CostApiGranularity.NONE,
            include_zero_costs: bool = False,
    ) -> CostApiItem:
        """
        Use the StackIT Cost API to get costs for a customer account.
        If called with a granularity, returns a CostApiItemWithDetails
        is returned. Otherwise, no details will be present.
        """
        url = f"{STACKIT_API_BASE_URL}/costs/{customer_account_id}/projects/{project_id}"
        params = {
            'from': from_date.strftime('%Y-%m-%d'),
            'to': to_date.strftime('%Y-%m-%d'),
            'depth': depth.value,
            'granularity': granularity.value,
            'includeZeroCosts': include_zero_costs,
        }
        bearer_token = self._auth.get_bearer_token()
        try:
            response = requests.get(
                url,
                params=params,
                headers={
                    'Authorization': f'Bearer {bearer_token}',
                    'Content-Type': 'application/json'
                }
            )
            response.raise_for_status()
        except Exception as e:
            raise CostApiException(f"GET {url} failed: {e}")
        data = response.json()
        if granularity == CostApiGranularity.NONE:
            try:
                return CostApiItem(**data)
            except Exception as e:
                raise CostApiException(
                    f"Failed to parse response as CostApiItem: {e}\n"
                    f"Response: {response.text}"
                )
        else:
            try:
                return CostApiItemWithDetails(**data)
            except Exception as e:
                raise CostApiException(
                    f"Failed to parse response as CostApiItemWithDetails: {e}\n"
                    f"Response: {response.text}"
                )



#!/usr/bin/python3
from enum import Enum
from pathlib import Path
from sys import exit
import argparse
from datetime import datetime, timedelta, timezone
from typing import NoReturn

from pydantic import BaseModel

from stackit_cost_monitoring.cost_api import CostApi, CostApiGranularity, CostApiItemWithDetails, CostApiException, \
    CostApiDepth
from stackit_cost_monitoring.auth import Auth, AuthException

SECONDS_PER_DAY = 24 * 3600
CENTS_PER_EURO = 100

DEFAULT_WARNING_EUROS = 10.0
DEFAULT_CRITICAL_EUROS = 50.0
DEFAULT_SA_KEY_JSON = Path.home() / ".stackit" / "sa-key.json"


class NagiosExitCodes(Enum):
    OK = 0
    WARNING = 1
    CRITICAL = 2
    UNKNOWN = 3


class ParsedArguments(BaseModel):
    customer_account_id: str
    project_id: str
    warning: float
    critical: float
    sa_key_json: Path
    skip_discounts: bool


class NagiosReporter:
    def __init__(self, args: ParsedArguments):
        self.args = args
        self.today = datetime.now(timezone.utc).date()
        self.yesterday = self.today - timedelta(days=1)
        self.today_cost = 0.0
        self.today_discounted_cost = 0.0
        self.yesterday_cost = 0.0
        self.yesterday_discounted_cost = 0.0

    def book_cost_item(self, cost_item: CostApiItemWithDetails):
        for report_data in cost_item.reportData:
            if report_data.timePeriod.start == self.today:
                self.today_cost += report_data.charge / CENTS_PER_EURO
                self.today_discounted_cost += report_data.discount / CENTS_PER_EURO
            elif report_data.timePeriod.start == self.yesterday:
                self.yesterday_cost += report_data.charge / CENTS_PER_EURO
                self.yesterday_discounted_cost += report_data.discount / CENTS_PER_EURO
            else:
                raise Exception(f"CostApi returned unexpected date: {report_data.timePeriod.start}")

    def do_report(self) -> NoReturn:
        today_cost = self.today_cost
        yesterday_cost = self.yesterday_cost
        """"
            StackIt's answer to ticket SSD-13595:
            
                The totalCharge value in the API response already includes all granted discounts.
                
            To detect pathological effects we should add the discounted costs to get an
            alarm before all our free budget has been used. By default we add the discounts.
        """
        if not self.args.skip_discounts:
            today_cost += self.today_discounted_cost
            yesterday_cost += self.yesterday_discounted_cost
        total_cost = max(today_cost, yesterday_cost)
        if total_cost >= self.args.critical:
            exit_code = NagiosExitCodes.CRITICAL
            message = f"Daily costs {total_cost:.2f} EUR >= {self.args.critical} EUR"
        elif total_cost >= self.args.warning:
            exit_code = NagiosExitCodes.WARNING
            message = f"Daily costs {total_cost:.2f} EUR >= {self.args.warning} EUR"
        else:
            exit_code = NagiosExitCodes.OK
            message = f"Daily costs {total_cost:.2f} EUR"
        return self._finish(exit_code, message)

    def _finish(self, status: NagiosExitCodes, message: str) -> NoReturn:
        warning = self.args.warning
        critical = self.args.critical
        perf_data_items = [
            f"yesterday_cost={self.yesterday_cost:.2f};{warning:.2f};{critical:.2f};0",
            f"yesterday_discounted_cost={self.yesterday_discounted_cost:.2f};{warning:.2f};{critical:.2f};0",
            f"today_cost={self.today_cost:.2f};{warning:.2f};{critical:.2f};0",
            f"today_discounted_cost={self.today_discounted_cost:.2f};{warning:.2f};{critical:.2f};0",
        ]
        perf_data = ' '.join(perf_data_items)
        print(f"{status.name}: {message} | {perf_data}")
        return exit(status.value)


def main():
    try:
        args = get_arguments()
        cost_item = get_cost(args)
        reporter = NagiosReporter(args)
        reporter.book_cost_item(cost_item)
        reporter.do_report()
    except (AuthException, CostApiException) as e:
        print(f"{NagiosExitCodes.UNKNOWN.name}: {e} |")
        exit(NagiosExitCodes.UNKNOWN.value)


def get_arguments() -> ParsedArguments:
    parser = argparse.ArgumentParser(
        description='Nagios plugin to monitor StackIT costs. The higher value '
        'of the cost of the present day (always 0?) and yesterday is used.',
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        '--customer-account-id',
        required=True,
        help='StackIT customer account ID'
    )
    parser.add_argument(
        '--project-id',
        required=True,
        help='StackIT project ID'
    )
    parser.add_argument(
        '-w', '--warning',
        type=float,
        default=DEFAULT_WARNING_EUROS,
        help=f"Warning threshold for 24h cost in EUR (default: {DEFAULT_WARNING_EUROS:.2f})"
    )
    parser.add_argument(
        '-c', '--critical',
        type=float,
        default=DEFAULT_CRITICAL_EUROS,
        help=f"Critical threshold for 24h cost in EUR (default: {DEFAULT_CRITICAL_EUROS:.2f})"
    )
    parser.add_argument(
        '--sa-key-json',
        type=Path,
        default=DEFAULT_SA_KEY_JSON,
        help=f"Path to StackIT credentials in JSON format (default: {DEFAULT_SA_KEY_JSON})"
    )
    parser.add_argument(
        '--skip-discounts',
        action='store_true',
        help='Skip discounted costs in calculation.'
    )

    parsed_arguments = ParsedArguments(**parser.parse_args().__dict__)
    if parsed_arguments.warning < 0.0:
        raise ValueError("Warning threshold must be >= 0.0")
    if parsed_arguments.critical <= parsed_arguments.warning:
        raise ValueError("Critical threshold must be > warning threshold")
    return parsed_arguments


def get_cost(args) -> CostApiItemWithDetails:
    auth = Auth(args.sa_key_json)
    cost_api = CostApi(auth)
    today = datetime.now(timezone.utc).date()  # StackIT, ticket SSD-13595: UTC is used
    yesterday = today - timedelta(days=1)

    result = cost_api.get_project_costs(
        args.customer_account_id,
        args.project_id,
        from_date=yesterday,
        to_date=today,
        granularity=CostApiGranularity.DAILY,
        depth=CostApiDepth.PROJECT,
        include_zero_costs=False,
    )

    if not isinstance(result, CostApiItemWithDetails):
        raise Exception("CostAPI returned item without report data!")
    return result


if __name__ == '__main__':
    main()



from dataclasses import dataclass
from datetime import datetime, timedelta, timezone

import json
from pathlib import Path
import uuid
from typing import Optional

import jwt
import requests
from pydantic import BaseModel

AUTH_CACHE_EXPIRATION = timedelta(minutes=10)
STACKIT_TOKEN_URL = 'https://service-account.api.stackit.cloud/token'


class _StackItRawCredentials(BaseModel):
    iss: str
    sub: str
    aud: str
    kid: str
    private_key: str


@dataclass
class _StackItBearerCredentials:
    token: str
    expiry: datetime


class AuthException(Exception):
    pass


class Auth:
    def __init__(self, sa_key_json_path: Path):
        self.sa_key_json_path = sa_key_json_path
        self._raw_credential_cache: Optional[_StackItRawCredentials] = None
        self._token_cache: Optional[_StackItBearerCredentials] = None

    def _get_raw_credentials(self) -> _StackItRawCredentials:
        try:
            if self._raw_credential_cache is None:
                with open(self.sa_key_json_path, 'r') as f:
                    credentials_data = json.load(f)['credentials']
                    self._raw_credential_cache = _StackItRawCredentials(
                        iss=credentials_data['iss'],
                        sub=credentials_data['sub'],
                        aud=credentials_data['aud'],
                        kid=credentials_data['kid'],
                        private_key=credentials_data['privateKey']
                    )
            return self._raw_credential_cache
        except Exception as e:
            error_context = f"Failed to load credentials from {self.sa_key_json_path}"
            raise AuthException(f"{error_context}: {e}")

    def _generate_jwt_token(self) -> str:
        credentials = self._get_raw_credentials()
        now = datetime.now(timezone.utc)

        payload = {
            'iss': credentials.iss,
            'sub': credentials.sub,
            'aud': credentials.aud,
            'iat': now,
            'exp': now + AUTH_CACHE_EXPIRATION,
            'jti': str(uuid.uuid4())
        }

        token = jwt.encode(
            payload,
            credentials.private_key,
            algorithm='RS512',
            headers={'kid': credentials.kid}
        )
        return token

    def _get_token_data(self) -> _StackItBearerCredentials:
        error_context = f"Failed to acquire bearer token from {STACKIT_TOKEN_URL}"
        jwt_token = self._generate_jwt_token()
        try:
            response = requests.post(
                STACKIT_TOKEN_URL,
                data={
                    'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
                    'assertion': jwt_token
                },
                headers={'Content-Type': 'application/x-www-form-urlencoded'}
            )
            response.raise_for_status()
        except Exception as e:
            raise AuthException(f"{error_context}: {e}")
        return _StackItBearerCredentials(
            token=response.json()['access_token'],
            expiry=datetime.now(timezone.utc) + AUTH_CACHE_EXPIRATION
        )

    def get_bearer_token(self) -> str:
        """Get a bearer token from StackIT."""
        if self._token_cache is None or self._token_cache.expiry < datetime.now(timezone.utc):
            self._token_cache = self._get_token_data()
        return self._token_cache.token
